#include <jni.h>
#include <stdlib.h>
#include "GL4D/gl4du.h"

JNIEXPORT void JNICALL Java_com_android_gles3jni_GLES3JNILib_init(JNIEnv* env, jobject obj);
JNIEXPORT int  JNICALL Java_com_android_gles3jni_GLES3JNILib_initTexture(JNIEnv * env, jobject obj, jint pw, jint ph);
JNIEXPORT void JNICALL Java_com_android_gles3jni_GLES3JNILib_resize(JNIEnv* env, jobject obj, jint width, jint height);
JNIEXPORT void JNICALL Java_com_android_gles3jni_GLES3JNILib_step(JNIEnv* env, jobject obj);

static GLuint _quad = 0, _tex[1] = { 0 }, _pIdVisu = 0;
static GLfloat _wratio = 1.0f, _tratio = 1.0f;
static GLint _vp[4];

static void quit(void) {
    if(_quad)
        gl4dgDelete(_quad);
    _quad = 0;
    if(_tex[0])
        glDeleteTextures(sizeof _tex / sizeof *_tex, _tex);
    _tex[0] = 0;
    if(_pIdVisu)
        gl4duDeleteProgram(_pIdVisu);
    _pIdVisu = 0;
}

JNIEXPORT void JNICALL
Java_com_android_gles3jni_GLES3JNILib_init(JNIEnv* env, jobject obj) {
    GL4D_GL_ERROR;
    char * argv[1] = {"GL4DES3"};
    if(gl4duHasInit())
        gl4duClean(GL4DU_ALL);
    gl4duInit(1, argv);
    GL4D_GL_ERROR;
    if(_quad == 0) {
        _quad = gl4dgGenQuadf();
        _pIdVisu  = gl4duCreateProgram(
                "<imvs>basic_inv.vs</imvs>"
                        "#version 300 es\n"
                        "precision mediump float;\n"
                        "layout (location = 0) in vec3 vsiPosition;\n"
                        "layout (location = 1) in vec3 vsiNormal;\n"
                        "layout (location = 2) in vec2 vsiTexCoord;\n"
                        "out vec2 vsoTexCoord;\n"
                        "void main(void) { \n"
                        "  gl_Position = vec4(vsiPosition, 1.0); \n"
                        "  vsoTexCoord = vec2(vsiTexCoord.x, 1.0 - vsiTexCoord.y); \n"
                        "}\n",
                "<imfs>tex.fs</imfs>"
                        "#version 300 es\n"
                        "precision mediump float;\n"
                        "uniform sampler2D tex;\n"
                        "in  vec2 vsoTexCoord;\n"
                        "out vec4 fragColor;\n"
                        "void main(void) { \n"
                        "  fragColor = texture(tex, vsoTexCoord).rgba;\n"
                        "}\n", NULL);
        gl4duAtExit(quit);
    }
    glClearColor(0.35, 0, 0, 1);
    glDisable(GL_DEPTH_TEST);
    GL4D_GL_ERROR;
}

JNIEXPORT int JNICALL
Java_com_android_gles3jni_GLES3JNILib_initTexture(JNIEnv * env, jobject obj, jint pw, jint ph) {
    if(_tex[0] == 0) {
        glGenTextures(sizeof _tex / sizeof *_tex, _tex);
    }
    glBindTexture(GL_TEXTURE_2D, _tex[0]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, pw, ph, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    _tratio = ph / (GLfloat)pw;
    return _tex[0];
}


JNIEXPORT void JNICALL
Java_com_android_gles3jni_GLES3JNILib_resize(JNIEnv* env, jobject obj, jint width, jint height) {
    int w, h;
    _wratio = height / (GLfloat)width;
    if(_wratio < _tratio) {
        h = height;
        w = h / _tratio;
    } else {
        w = width;
        h = w * _tratio;
    }
    _vp[0] = (width - w) / 2;
    _vp[1] = (height - h) / 2;
    _vp[2] = w;
    _vp[3] = h;
    glViewport(_vp[0], _vp[1], _vp[2], _vp[3]);
}

JNIEXPORT void JNICALL
Java_com_android_gles3jni_GLES3JNILib_step(JNIEnv* env, jobject obj) {
    glClear(GL_COLOR_BUFFER_BIT);
    glUseProgram(_pIdVisu);
    glActiveTexture(GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(_pIdVisu, "tex"), 0);
    glBindTexture(GL_TEXTURE_2D, _tex[0]);
    gl4dgDraw(_quad);
    glUseProgram(0);
    gl4duPrintFPS(stdout);
}

